<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Facades\Session;
use GuzzleHttp\Client;
use thiagoalessio\TesseractOCR\TesseractOCR;

class ChatbotController extends Controller
{
    private $apiKey;
    private $model;

    public function __construct()
    {
        $this->apiKey = env('AI_API_KEY');
        $this->model = 'gemini-1.5-flash-latest';
    }

    public function index()
    {
        $userName = Auth::user()->name;
        return view('chatbot.index', compact('userName'));
    }

    public function send(Request $request)
    {
        $userMessage = $request->input('message');
        $botResponse = $this->getBotResponse($userMessage);

        return response()->json(['message' => $botResponse]);
    }

    private function getBotResponse($message)
    {
        $url = "https://generativelanguage.googleapis.com/v1beta/models/{$this->model}:generateContent?key={$this->apiKey}";

        $chat = $this->getChatHistory();
        $chat[] = [
            'role' => 'user',
            'parts' => [['text' => $message]]
        ];

        $data = [
            'contents' => $chat,
            'generationConfig' => [
                'temperature' => 1,
                'topP' => 0.95,
                'topK' => 64,
                'maxOutputTokens' => 8192,
            ],
        ];

        $response = Http::post($url, $data);

        if ($response->successful()) {
            $result = $response->json();
            if (isset($result['candidates'][0]['content']['parts'][0]['text'])) {
                $botResponse = $result['candidates'][0]['content']['parts'][0]['text'];
                $chat[] = [
                    'role' => 'model',
                    'parts' => [['text' => $botResponse]]
                ];
                $this->saveChatHistory($chat);
                return $botResponse;
            }
        }

        return "Sorry, I couldn't process that request. Error: {$response->status()}, {$response->body()}";
    }

    private function getChatHistory()
    {
        if (!Session::has('chat_history')) {
            $initialChat = $this->initializeChat();
            Session::put('chat_history', $initialChat);
        }
        return Session::get('chat_history');
    }

    private function saveChatHistory($chat)
    {
        Session::put('chat_history', $chat);
    }

    private function initializeChat()
    {
        return [
            [
                'role' => 'user',
                'parts' => [[
                    'text' => "You are a helpful and friendly assistant for Crownscope Insurance Brokers, specializing in golfers' claims. Your job is to guide users through the insurance claims process. Always stay in context, be cheerful, and keep responses short and engaging. You will handle claims for five covers:

                                Loss or damage to golf equipment up to Ksh 150,000
                                Loss or damage to personal effects up to Ksh 50,000
                                Third-party liability up to Ksh 500,000
                                Personal accident up to Ksh 250,000
                                Hole-in-one up to Ksh 30,000

                                Always suggest what the user can do next. Follow these steps for each type of claim:

                                Golf Equipment:
                                1. Ask what was lost or damaged and wait for feedback.
                                2. List the items mentioned and ask if there's more. Repeat until they say there's no more.
                                3. Confirm the list and request receipts or photo proof for each item.
                                4. Go through the list and ensure all photos are submitted for each item.

                                Personal Effects:
                                1. Ask what was lost or damaged and wait for feedback.
                                2. List the items mentioned and ask if there's more. Repeat until they say there's no more.
                                3. Request a photo or scanned document of police reports for each item.
                                4. Go through the list and ensure all photos or documents are submitted for each item.
                                5. Ask for a video of them explaining WHat happenned max 2 mins giving details of how where and any other necessary.

                                Third-Party Liability:
                                1. Ask for details about the incident and wait for feedback.
                                2. Request any supporting documents or evidence related to the incident.
                                3. Ask for a video of them explaining WHat happenned max 2 mins giving details of how where and any other necessary.

                                Personal Accident:
                                1. Express sympathy and ask what happened.
                                2. Request a clear image of the hospital receipt.
                                3. Ask for a video of them explaining WHat happenned max 2 mins giving details of how where and any other necessary.

                                Hole-in-One:
                                1. Be excited and ask how they feel.
                                2. Request receipts of any party they attended.
                                3. Ask for a video of them explaining WHat happenned max 2 mins giving details of how where and any other necessary.

                                In each case, always thank the user for filing the claim."
                ]]
            ],
            [
                'role' => 'model',
                'parts' => [['text' => "Understood! I'm ready to assist users with their golf insurance claims for Crownscope Insurance Brokers. I'll be friendly, engaging, and follow the specified steps for each type of claim. How can I help you today?"]]
            ]
        ];
    }

    public function upload(Request $request)
    {
        $request->validate([
            'file' => 'required|mimes:jpg,jpeg,png,bmp,tiff|max:2048',
        ]);

        $file = $request->file('file');
        $fileName = uniqid() . '_' . time() . '.' . $file->getClientOriginalExtension();
        $filePath = public_path('uploads/' . $fileName);
        $file->move(public_path('uploads'), $fileName);

        try {
            $fileRead = (new TesseractOCR($filePath))->lang('eng')->run();
            $botResponse = $this->getBotResponse($fileRead);

            return response()->json([
                'message' => $botResponse,
                'uploaded_image_url' => asset('uploads/' . $fileName)
            ]);
        } catch (\Exception $e) {
            return response()->json(['error' => $e->getMessage()], 500);
        }
    }
}